-- Auras Handlers module (C) 2011 Michael

local Addon= BurnIt

local GetTime= GetTime
local GetSpellInfo= GetSpellInfo
local UnitAura= UnitAura
local UnitGUID= UnitGUID
local GetCombatRatingBonus= GetCombatRatingBonus
local GetMastery= GetMastery
local fmt= string.format
local select= select
local next = next
local unpack = unpack
local tinsert = table.insert
local tremove = table.remove
local floor = math.floor
local bit_band= bit.band

--{{ Common functions used by Aura Handlers

local function CreateBar(text, icon, dbx, description)
	local bar= Addon.CreateBar(text,icon)
	bar.dbx= dbx
	if description then 
		bar.description= description 
	end
	if dbx.hasStack then 
		bar:AddStack() 
	end	
	return bar
end

local function UpdateShared(self, aura, ...)
	local bar = self.bar
	local count,_, duration, expiration = select(4,UnitAura(aura.unitId, select(13,...) or self.spellName, nil , aura.auraMask))
	if duration then
		bar:SetText(bar.defaultText)
		bar:SetTime(expiration,duration)
		bar:SetStack(count)
		bar:Enable()
	else
		bar:Disable()
	end
end

local function DotsShared(self, aura, amount, critical)
	local bar= self.bar
	if bar.enabled then
		if critical then
			bar:SetText(amount.."!")
		else
			bar:SetText(amount)
		end
	end	
end	

local function CreateAuraShared(self, setup) 
	if setup.dots and (not self.Dots) then
		self.Dots= DotsShared 
	end	
	local spellName,_,spellIcon= GetSpellInfo( setup.spells[1] )
	self.spellName= spellName
	self.bar= CreateBar(spellName,spellIcon, setup.bars[1])
	return self
end

--}}

--{{ Handler for normal auras on target/focus/etc, more than one aura can be provided (but the auras cannot be actived at same time on the same target)
local CreateAuraHandler
do
	CreateAuraHandler= function(setup)
		return CreateAuraShared( { Update= UpdateShared }, setup )
	end
end
--}}

--{{ Handler for multiple auras: tracks several auras on the same bar, the bar shows the aura with greater expiration time
local CreateMultipleAuraHandler
do
	local function Update(self, aura)
		local duration,expiration,index
		local spells= self.spellNames
		for i=#spells,1,-1 do
			local d,e = select(6,UnitAura(aura.unitId, spells[i], nil, aura.auraMask))
			if e and (not expiration or e>expiration) then
				duration,expiration,index= d,e,i
			end
		end
		if duration then
			local bar,spellName= self.bar, self.spellNames[index]
			if spellName ~= self.curSpellName then
				self.curSpellName= spellName
				bar:SetText( spellName )
				bar:SetIcon( self.spellIcons[index] )		
			end
			bar:SetTime(expiration,duration)
			bar:Enable()
		else
			self.curSpellName=nil
			self.bar:Disable()
		end
	end
	CreateMultipleAuraHandler= function(setup)
		local self= { Update= Update, Dots= setup.dots and DotsShared or nil }
		local spellName,spellIcon
		self.spellNames= {}
		self.spellIcons= {}
		local spells= setup.spells
		for i=#spells,1,-1 do
			spellName,_,spellIcon= GetSpellInfo(spells[i])
			tinsert(self.spellNames,spellName)
			tinsert(self.spellIcons,spellIcon)
		end
		self.bar= CreateBar(spellName,spellIcon, setup.bars[1])
		return self
	end
end
--}}

--{{ Handler for Mirror image, or similar auras (Aura is hidden so aura duration must be provided to the handler)
local CreateMirrorImageHandler
do
	local function Add(self, aura)
		self.bar:SetTime(GetTime()+self.duration, self.duration)
		self.bar:Enable()
	end
	CreateMirrorImageHandler= function(setup)
		return CreateAuraShared( { Add=Add, duration=setup.duration }, setup )
	end
end
--}}

--{{ Handler for Living bombs: Tracks target living bomb, and create several bars for non target living bombs
local CreateLivingBombHandler
do
	local function AssignBar(self, destGUID)
		local bar = self.bars[destGUID]
		if not bar then
			bar = tremove(self.unusedBars)
			if not bar then
				for g,b in next,self.bars do
					if not b.enabled then
						bar = b
						self.bars[g] = nil
						break
					end
				end
			end
			self.bars[destGUID] = bar
		end
		return bar
	end
	local function UpdateTarget(self, aura)
		local duration, expiration = select(6,UnitAura("target", self.spellName, nil , aura.auraMask))
		if duration then
			self.bar:SetTime(expiration,duration)
			self.bar:Enable()
		else
			self.bar:Disable()
		end
		return duration, expiration
	end
	local function Update(self,aura)
		UpdateTarget(self,aura)
		local targetGUID= UnitGUID("target")
		for guid,bar in next,self.bars do
			bar:SetAlpha( guid==targetGUID and 1 or self.alpha)
		end
	end
	local function Add(self,aura,...)
		local duration,expiration
		local destGUID= select(8,...)
		-- Update target living bomb
		local isTarget= (destGUID==UnitGUID("target"))
		if isTarget then
			duration,expiration= UpdateTarget(self,aura)
		elseif destGUID==UnitGUID("mouseover") then
			duration, expiration = select(6,UnitAura("mouseover", self.spellName, nil, aura.auraMask))
		elseif destGUID==UnitGUID("focus") then
			duration, expiration = select(6,UnitAura("focus", self.spellName, nil, aura.auraMask))
		end
		-- Calculate duration by Hand
		if not duration then 
		    -- local haste = ((GetCombatRatingBonus(20)/100) + 1) -- TODO calculate living bomb duration
			duration= 12
			expiration= GetTime()+duration
		end
		-- Enabled nontarget living bomb
		local bar = AssignBar(self, destGUID)
		if bar then
			bar:SetTime(expiration,duration)
			bar:SetText(select(9,...))
			bar:Enable()
			bar:SetAlpha( isTarget and 1 or self.alpha )
		else
			print("BurnIt Error: no bar available, lost living bomb tracking")
		end	
	end
	local function Remove(self,aura,...)
		local destGUID= select(8,...)
		-- Disabling target living bomb
		if UnitGUID("target")==destGUID then
			self.bar:Disable()
		end
		-- Disabling nontarget living bomb
		local bar= self.bars[destGUID]
		if bar then
			self.bars[destGUID]= nil
			tinsert(self.unusedBars, bar)
			bar:Disable()
		end
	end
	local function Dots(self, aura, amount, critical , ...)
		if UnitGUID("target")==select(8,...) then
			DotsShared(self,aura,amount,critical)
		end
	end
	CreateLivingBombHandler= function(setup)
		local self={ Update=Update, Add=Add, Remove=Remove, Dots= Dots}
		local spellName,_,spellIcon= GetSpellInfo( setup.spells[1] )
		self.spellName= spellName
		self.color= setup.bars[1].color
		self.alpha= setup.alpha or 0.6
		self.bars= {}
		self.unusedBars= {}
		self.bar          = CreateBar(spellName, spellIcon, setup.bars[1]) -- target living bomb
		self.unusedBars[1]= CreateBar(spellName, spellIcon, setup.bars[2], spellName.." 1" )
		self.unusedBars[2]= CreateBar(spellName, spellIcon, setup.bars[3], spellName.." 2" )
		self.unusedBars[3]= CreateBar(spellName, spellIcon, setup.bars[4], spellName.." 3" )
		self.unusedBars[4]= CreateBar(spellName, spellIcon, setup.bars[5], spellName.." 4" )
		return self
	end
end
--}}

--{{ Handler for Cauterize debuff and cooldown or similar auras
local CreateCauterizeHandler
do
	local function Add(self,aura,params)
		local duration, expiration = select(6,UnitAura(aura.unitId, self.spellName, nil, aura.auraMask))
		if duration then
			self.cooldownStart= expiration-duration
			local bar= self.barAura
			bar:SetTime(expiration,duration)
			bar:Enable()
		end	
	end
	local function Remove(self,aura,params)
		self.barAura:Disable()
		local bar= self.barCooldown
		bar:SetTime(self.cooldownStart+self.cooldown,self.cooldown)
		bar:Enable()
	end
	CreateCauterizeHandler= function(setup)
		local self= { Add=Add, Remove=Remove, cooldown= setup.cooldown } 
		local spellName,_,spellIcon= GetSpellInfo( setup.spells[1] )
		self.spellName  = spellName
		self.barAura    = CreateBar(spellName, spellIcon, setup.bars[1])
		self.barCooldown= CreateBar(spellName, spellIcon, setup.bars[2], spellName .. " cooldown")
		return self
	end
end
--}}

--{{ Handler for Polymorph, or similar auras 
local CreatePolymorphHandler
do
	local function Add(self, aura, ...)
		self.bar:SetTime(GetTime()+self.duration, self.duration)
		self.bar:SetText(select(9,...))
		self.bar:Enable()
	end
	local function Remove(self, aura)
		self.bar:Disable()
	end
	CreatePolymorphHandler= function(setup)
		return CreateAuraShared( { Add=Add, Remove=Remove, duration=setup.duration}, setup )
	end
end
--}}

--{{ Handler for Polymorph, or similar auras 
local CreateFlameStrikeHandler
do
	local function Add(self, aura, ...)
		self.bar:SetTime(GetTime()+self.duration, self.duration)
		self.bar:Enable()
	end
	CreateFlameStrikeHandler= function(setup)
		local self= { Add=Add, duration=setup.duration }
		local spellName,_,spellIcon= GetSpellInfo( setup.spells[1] )
		self.spellName= spellName
		self.bar= CreateBar( spellName .. (setup.suffix or ""), spellIcon, setup.bars[1])
		return self
	end
end
--}}

--{{ Handler for Ignite
local CreateIgniteHandler
do
	--{{ Special code to calculate ignite munching
	local realIgniteAmount = 0
	local specIgniteAmount = 0
	local specIgniteTick   = 0
	local specPredCount    = 0
	local specPredFail     = 0
	local function InitMunchingReporting()
		local eventFrame = CreateFrame("Frame", nil, Addon.MainFrame)
		eventFrame:SetScript("OnEvent", function(self, event) 
			if event=="PLAYER_REGEN_DISABLED" then
				realIgniteAmount, specIgniteAmount, specIgniteTick, specPredCount, specPredFail = 0,0,0,0,0
			elseif Addon.MageSpec=="fire" then
				if specIgniteAmount>0 then
					local diff= specIgniteAmount - realIgniteAmount
					if math.abs(diff)<1000 then 
						specIgniteAmount, diff = realIgniteAmount, 0
					end
					local perc= diff / specIgniteAmount * 100
					print( string.format("|c0000d0ffBurnIt Ignite Damage.  Expected:|r |c0000ff00 %d |r|c0000d0ff Real: |r|c00ffff00 %d |r|c0000d0ff Munched: |r|c00ff0000 %d (%d%%) |r ", specIgniteAmount, realIgniteAmount, diff, perc  ) )
				end
				if specPredCount>0 then
					local perc = specPredFail / specPredCount * 100
					print( string.format("|c0000d0ffBurnIt Ignite Predictions.  Total:|r |c0000ff00 %d |r|c0000d0ff Fails: |r|c00ff0000 %d (%d%%) |r", specPredCount, specPredFail, perc  ) )
				end
			end
		end)
		eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
		eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
	end
	--}}
	
	local igniteGUID
	local igniteDmg        = 0
	local igniteDmgNew     = 0
	local igniteExpiration = 0
	local ingniteFreeTick  = false
	local igniteSpells  = {
		[133]   = true, -- Bola de fuego
		[2948]  = true, -- Agostar
		[2120]  = true, -- Fogonazo
		[2136]  = true, -- Explosion de fuego
		[11113] = true, -- Ola explosiva
		[11366] = true, -- Piro
		[92315] = true, -- Piro!
	}
	
	-- Combat Log Event Periodic Damage tracking
	local function Dots(self, aura, amount, critical)
		local bar = self.bar
		if bar.enabled then
			local expiration = select(7, UnitAura(aura.unitId, self.spellName, nil , aura.auraMask))
			local duration   = (expiration or 0) - GetTime()
			if duration>2.25 then
				igniteDmg = amount * 2
			elseif duration>0.25 then
				igniteDmg = amount
			else
				igniteDmg = 0
			end
			if igniteFreeTick then 				-- Special case, due to an ignite bug tick damage is not removed
				igniteDmg = igniteDmg + amount 	-- when new ignite damage exists (target received a critical spell recently)
				igniteFreeTick = false
			end
			bar:SetText( amount )
			--
			realIgniteAmount = realIgniteAmount + amount
			if specIgniteTick>0 then
				local diff = amount - specIgniteTick
				if math.abs(diff)>500 then
					specPredFail = specPredFail + 1
					-- print("BurnIt predicion failed. Espected:", specIgniteTick, "Real:", amount, "Diff:", diff)
				end
				specPredCount = specPredCount + 1
			end
		end	
	end
	
	-- Combat Log Event Apply&Refresh Aura tracking (called when target changed too)
	local function Update(self, aura , ...)
		local duration
		local bar  = self.bar
		local guid = UnitGUID(aura.unitId) 
		if guid ~= igniteGUID then
			igniteGUID, igniteFreeTick, igniteDmg, igniteDmgNew, igniteExpiration = guid, false, 0, 0, 0
		end
		duration, igniteExpiration = select(6,UnitAura(aura.unitId, self.spellName, nil , aura.auraMask))
		if duration then
			igniteDmg       = igniteDmg + igniteDmgNew
			igniteDmgNew    = 0
			local tickCount = select(2,...)=="SPELL_AURA_REFRESH" and 3 or 2
			local tickDmg   = floor( igniteDmg/tickCount )
			bar:SetText( tickDmg>0 and "~"..tickDmg or bar.defaultText )
			bar:SetTime( igniteExpiration, duration )
			bar:Enable()
			specIgniteTick = tickDmg
			-- print("Prediction: ", tickDmg, " Ticks:", tickCount, "Duration:", duration)
		else
			igniteFreeTick, igniteDmg, igniteDmgNew, igniteExpiration = false, 0, 0, 0
			bar:Disable()
		end
	end
	
	-- Combat log events Fire Crits tracking
	local function Damage(self, aura, ...) 
		if  select( 21, ... ) and 										 -- Critical spell ?	
			igniteSpells[ select(12,...) ] and							 -- Valid fire spell ?
			aura.sourceGUID==select(4,...) and 						     -- casted by me ?
			bit_band(select(10,...),aura.destFlags)==aura.destFlags then -- is target ?			
				local remain     = igniteExpiration - GetTime() - 0.1 	 -- asuming 100 ms latency
				igniteFreeTick   = remain>4 or (remain>2 and remain<2.5) or (remain>0 and remain<0.5)
				igniteDmgNew     = select( 15, ... ) * 0.4 * (GetMastery()*0.028+1)
				specIgniteAmount = specIgniteAmount + igniteDmgNew
		end
	end

	CreateIgniteHandler= function(setup)
		local self= CreateAuraShared( { Damage= Damage, Update= Update, Dots= Dots }, setup )
		self.bar:AddSpark(2)
		self.bar:AddSpark(4)
		InitMunchingReporting()
		return self
	end
end

--{{ Handler for Arcane Blast
local CreateArcaneBlastHandler
do
	local function Damage(self, aura, ...)
		if aura.sourceGUID==select(4,...) and select(12,...)==30451 and self.bar.stackCount==4 then
			UpdateShared(self, aura, ...)
		end
	end
	CreateArcaneBlastHandler= function(setup)
		return CreateAuraShared( { Update= UpdateShared, Damage= Damage }, setup )
	end
end
--}}

--{{{{ Publishing Aura Handlers

Addon.AuraHandlers= {
 ["CommonAura"]  = CreateAuraHandler,
 ["MultipleAura"]= CreateMultipleAuraHandler,
 ["LivingBomb"]  = CreateLivingBombHandler,
 ["HiddenAura"]  = CreateMirrorImageHandler,
 ["CooldownAura"]= CreateCauterizeHandler,
 ["Polymorph"]   = CreatePolymorphHandler,
 ["FlameStrike"] = CreateFlameStrikeHandler,
 ["IgniteAura"]  = CreateIgniteHandler,
 ["ArcaneBlastAura"] = CreateArcaneBlastHandler,
}

--}}}}